#ifndef _DDWRAPPER_HEADER
#define _DDWRAPPER_HEADER

#include <windows.h>
#include <ddraw.h>

// the structure which holds pixel RGB masks
typedef struct _RGBMASK
{
   unsigned long rgbRed;   // red component
   unsigned long rgbGreen; // green component
   unsigned long rgbBlue;  // blue component
} RGBMASK;

// the structure which holds screen surface info (5,6,5 or 5,5,5 and masking)
typedef struct _RGB16
{
   RGBQUAD depth;
   RGBQUAD Amount;
   RGBQUAD Position;
   RGBMASK Mask;
} RGB16;

// ### Macros Extracts Red, Green, and Blue Component, and create RGB value ---
#define RED16  (p)       (p >> rgb16.Position.rgbRed)

// ### Extracts Green Component ---
#define GREEN16(p)       ((p & rgb16.Mask.rgbGreen) >> rgb16.Position.rgbGreen)

// ### Extracts Blue Component ---
#define BLUE16 (p)       (p & rgb16.Mask.rgbBlue)

// ### Creates RGB Pixel Value ---
#define RGB16(r, g, b) ((r << rgb16.Position.rgbRed) | (g << rgb16.Position.rgbGreen) | b)

class DDWrapper
{
    public:
    	LPDIRECTDRAW        pDD;         // DirectDraw object
    	LPDIRECTDRAWSURFACE pDDSPrimary; // DirectDraw primary surface
    	LPDIRECTDRAWCLIPPER pDDClipper;  // Clipper for windowed mode
      	LPDIRECTDRAWPALETTE pDDPalette;
        PALETTEENTRY        ape[256];

    	HWND                ddWnd;       // Handle of window
    	BOOL                bFullScreen; // are we in fullscreen mode?
    	BOOL                bAnimating;  // are we animating?
    	BOOL	            dDrawActive; // determine if direct draw was initialised o.k.

    	int 		        bitDepth;    // Used in pixel plotting routines.
    	RGB16		        rgb16;	 	 // Used in 16bit pixel plotting.
        DDSURFACEDESC	    ddsd;        // Used in fast pixel plotting.
        int 				maxX, maxY;

    // *** Initialisation Methods ***
    int SetRGB16(LPDIRECTDRAWSURFACE Surface);
    int Create256Palette(void);
    int InitDDraw  (HWND hwnd, BOOL fullscreen, int width, int height, int depth);

    // *** Constructor ***
    DDWrapper ()
    {
    	pDD         = NULL;
    	pDDSPrimary = NULL;
    	pDDPalette  = NULL;
    	pDDClipper  = NULL;
    	bFullScreen = FALSE;
    	bAnimating  = FALSE;
    }

    // *** Destructors ***
    void CleanUp (void)
    {
    	// DirectDraw object clean out
    	if (pDD != NULL)
    	{
	    	pDD->Release();
    	    pDD = NULL;
    	}

		pDDSPrimary = NULL;
    	pDDPalette  = NULL;
		pDDClipper  = NULL;

		dDrawActive = false;
    }

    // Auto lock surface method
    inline void PlotPixel (int x, int y, COLORREF col);

    // **** Fast pixel plot routine ****
    inline BOOL LockPrimarySurface   (void);
    inline void FastPlotPixel        (int x, int y, COLORREF col);
    inline void UnlockPrimarySurface (void);

    ~DDWrapper ()
    {
    	CleanUp();
    }
};

#endif // _DDWRAPPER_HEADER
