#ifndef _FCLASSES_CODE
#define _FCLASSES_CODE

#include "fclasses.h"
#include "fixed_math.h"

int PixelSize  = 1;
int ExplodeSnd = 1;

// ############################################################################
// Constructors

Flare::Flare (void)
{
    xPos          = yPos = xPlot = yPlot = ground = gravty = thrust = 0;
    gndTerminated = alive = FALSE;
    gravConst     = (1 << SHIFT) >> (randomTab (3)+1);// gravity constant... this case it's randomTab.

    // trailer buffer settings.
    overBuf     = FALSE;
    bufPnt      = -1;
    trailLen    = 0;
    trailer     = NULL;
    colour      = RGB(255,255,255);
}

Flare::Flare (int  x,           int      y,
              int  thr,         int      grav,
              int  cyc,         int      gnd,
              int  trailLength, COLORREF col)
{
    init (x, y, thr, grav, cyc, gnd, trailLength, col);
}

void Flare::init (int  x,           int      y,
                  int  thr,         int     grav,
                  int  cyc,         int      gnd,
                  int  trailLength, COLORREF col)
{
    xPos   = x;
    yPos   = y;
    thrust = thr;
    ground = gnd-1;
    cycles = cyc;
    colour = col;
    gravty = grav;

    // init internals ***** Change XPos, yPos and assign to xPlot, yPlot *****
    xPlot  = yPlot = 0;
    gndTerminated = FALSE;           // flare hasn't hit the ground!
    alive         = TRUE;

    //**** Allocate trailer memory ****
    // If memory is already allocated, but trail length has changed. Then
    // delete/free memory.
    if ((trailer != NULL) && (trailLen != trailLength))
    {
        delete[] trailer;
        trailer = NULL;
    }

    // Allocate memory
    if (trailer == NULL)
       trailer = new SaveXY[trailLength];

    // buffer variables for trailer
    trailLen = trailLength;
    overBuf  = FALSE;
    bufPnt   = -1;
}

inline void Flare::removePoint (DDWrapper& ddObj, int x, int y)
{
    if (y > ground)
       y = ground;

    switch (PixelSize)
    {
        case 1:
             ddObj.FastPlotPixel(x, y, 0);
             break;

        case 2:
             ddObj.FastPlotPixel(x, y, 0);
             ddObj.FastPlotPixel(x, y+1, 0);
             break;

        case 3:
             ddObj.FastPlotPixel(x+1, y-1, 0);
             ddObj.FastPlotPixel(x, y, 0);
             ddObj.FastPlotPixel(x+1, y, 0);
             break;

        case 4:
             ddObj.FastPlotPixel(x, y, 0);
             ddObj.FastPlotPixel(x-1, y, 0);
             ddObj.FastPlotPixel(x, y-1, 0);
             ddObj.FastPlotPixel(x-1, y-1, 0);
    }

}

inline void Flare::showPoint (DDWrapper& ddObj, int x, int y)
{
    if (y > ground)
       y = ground;

    switch (PixelSize)
    {
        case 1:
             ddObj.FastPlotPixel(x, y, colour);
             break;

        case 2:
             ddObj.FastPlotPixel(x, y,   colour);
             ddObj.FastPlotPixel(x, y+1, colour);
             break;

        case 3:
             ddObj.FastPlotPixel(x+1, y-1, colour);
             ddObj.FastPlotPixel(x, y, colour);
             ddObj.FastPlotPixel(x+1, y, colour);
             break;

        case 4:
             ddObj.FastPlotPixel(x, y,     colour);
             ddObj.FastPlotPixel(x-1, y,   colour);
             ddObj.FastPlotPixel(x, y-1,   colour);
             ddObj.FastPlotPixel(x-1, y-1, colour);
    }

}

BOOL Flare::move (DDWrapper& ddObj)
{
    if (alive)
    {
        // update flare
        if (trailLen > 0)
        {
            // decrement cycles of life for firework
            cycles--;

            yPlot  = yPlot + gravty;      // decrease virtical lift
            xPlot  = xPlot + thrust;      // decrease horizontal thrust
            gravty = gravty - gravConst;  // update gravity pull

            // **** Trailer maintenace ****
            // load buffer from start
            if (overBuf == FALSE)
            {
                // If item exceeds buffer, rollover to start of buffer (queue system)
                if (bufPnt >= trailLen-1)
                {
                    bufPnt  = -1;
                    overBuf = TRUE;

                    removePoint (ddObj, trailer [bufPnt+1].x, trailer [bufPnt+1].y);
                }

                bufPnt++;
                trailer[bufPnt].x = xPos + (xPlot >> SHIFT);
                trailer[bufPnt].y = yPos - (yPlot >> SHIFT);
            }
            else // cycle thru queue buffer.
            {
                if (cycles <= trailLen)
                {
                    trailLen--;
                    removePoint (ddObj, trailer [trailLen].x, trailer [trailLen].y);
                }

                if (bufPnt >= trailLen - 1)
                   bufPnt = -1;

                bufPnt++;
                removePoint (ddObj, trailer [bufPnt].x, trailer [bufPnt].y);
                trailer [bufPnt].x = xPos + (xPlot >> SHIFT);
                trailer [bufPnt].y = yPos - (yPlot >> SHIFT);
            }

            showPoint (ddObj, trailer [bufPnt].x, trailer [bufPnt].y);
        }
        else
        {
            removePoint (ddObj, trailer [0].x, trailer [0].y);
            alive = FALSE;
        }
    }
    return alive;
}

// ############################################################################
// Destructor
Flare::~Flare(void)
{
	if (trailer != NULL)
    {
	    delete[] trailer;
        trailer = NULL;
    }
}


// ############################################################################
// ############################################################################
// ############################################################################

void SkyRocket::init (int  starBNum,
                      int  x,           int      y,
                      int  thr,         int      grav,
                      int  cyc,         int      gnd,
                      int  trailLength, COLORREF col)
{
    //**** Allocate explode memory ****
    // If memory is already allocated, but explode number has changed. Then
    // delete/free memory.
    if ((pStarBurst != NULL) && (starBurstNum != starBNum))
    {
        delete[] pStarBurst;
        pStarBurst = NULL;
    }

    if (pStarBurst == NULL)
       pStarBurst = new Flare[starBNum];

    // Initialise main rocket ....
    mainRocket.init (x, y, thr, grav, cyc, gnd, trailLength, col);

    // Initialise explode variables.
    multiBang    = -1;
    explode      = FALSE;
    starBurstNum = starBNum;
    colour       = col;
    trailLen     = trailLength;

    // Select the increment style according to current colour value.
    switch ((colItemChange = randomTab(3)))
    {
         // red
         case 0:
              if (GetRValue(colour) < 127)
                 colChangeAmnt = 1;
              else
                 colChangeAmnt = -1;
              break;

         // green
         case 1:
              if (GetGValue(colour) < 127)
                 colChangeAmnt = 1;
              else
                 colChangeAmnt = -1;
              break;


         // blue
         case 2:
             if (GetBValue(colour) < 127)
                colChangeAmnt = 1;
             else
                colChangeAmnt = -1;
             break;
    }
}

// ############################################################################

BOOL SkyRocket::move (DDWrapper& ddObj)
{
    // If no memory allocated .... leave.
    if (pStarBurst == NULL)
       return FALSE;

    // Update main rocket, and then explode!
    if (explode == FALSE)
    {
        if (mainRocket.move(ddObj) == FALSE)
        {
            // Set up explosion when main rocket dies.
            int x, y;
            explode = TRUE;

            // intialise star burst.
            mainRocket.getCurrentPos (x, y);

            // Implement random explosion type, two setup options exist.
            int count;

            if (multiBang < 0)
            	multiBang = randomTab(BOOMNUM);

            switch (multiBang)
            {
                case 0:
                     // Implement a super nova explosion.
                     for (count = 0; count < starBurstNum; count++)
                     {
                          pStarBurst[count].init (x,
                                                  y,
                                                 ((randomTab (60)-30) << SHIFT) >> randomTab(5),
                                                 ((randomTab (60)-30) << SHIFT) >> randomTab(5),
                                                 randomTab(100),
                                                 MAXY,
                                                 randomTab(trailLen)+5,
                                                 colour);
                     }
                     break;

                case 1:
                     for (count = 0; count < starBurstNum; count++)
                     {
                        pStarBurst[count].init (x + (randomTab (10) - 5),
                                                y + (randomTab (10) - 5),
                                                ((randomTab (30) -15) << SHIFT) >> (randomTab(2) + 2),
                                                ((randomTab (100)-10) << SHIFT) >> (randomTab(2) + 2),
                                                randomTab(100),
                                                MAXY,
                                                randomTab(trailLen) + 5,
                                                colour);
                     }
                     break;

                case 2:
                     for (count = 0; count < starBurstNum; count++)
                     {
                        pStarBurst[count].init (x + (randomTab (10) - 5),
                                                y + (randomTab (10) - 5),
                                                ((randomTab (100)-50) << SHIFT) >> (randomTab(2) + 2),
                                                ((randomTab (100)-10) << SHIFT) >> (randomTab(4) + 2),
                                                randomTab(100),
                                                MAXY,
                                                randomTab(trailLen) + 5,
                                                colour);
                     }
                     break;

                case 3:
                     // Circle small explosion.
                     for (count = 0; count < starBurstNum; count++)
                     {
                        pStarBurst[count].init (x + (randomTab (10) - 5),
                                                y + (randomTab (10) - 5),
                                                ((randomTab (20) - randomTab(20)) << SHIFT) >> (randomTab(2) + 1),
                                                ((randomTab (20) - randomTab(20)) << SHIFT) >> (randomTab(2) + 1),
                                                randomTab(100),
                                                MAXY,
                                                randomTab(trailLen) + 5,
                                                colour);
                     }
                     break;

                case 4:
                     {
                        // spiral explosion.
                        int r;
                        int xSin;
                        int yCos;

                        for (count = 0; count < starBurstNum; count++)
                        {
                            r    = randomTab(NUMOFDEGREES);
                            xSin = (int)((SIN (r) * randomTab(100)+10) >> SHIFT);
                            yCos = (int)((COS (r) * randomTab(100)+10) >> SHIFT);
                            pStarBurst[count].init (x + xSin,
                                                    y + yCos,
                                                    (yCos << SHIFT) >> 2,
                                                    (xSin << SHIFT) >> 2,
                                                    randomTab(100),
                                                    MAXY,
                                                    randomTab(trailLen) + 5,
                                                    colour);
                        }
                     }
                     break;

                case 5:
                     {
                        // ????????.
                        int r;
                        int xSin;
                        int yCos;

                        for (count = 0; count < starBurstNum; count++)
                        {
                            r    = randomTab(NUMOFDEGREES);
                            xSin = (int)((SIN (r) * randomTab(100)+10) >> SHIFT);
                            yCos = (int)((COS (r) * randomTab(100)+10) >> SHIFT);
                            pStarBurst[count].init (x,
                                                    y,
                                                    (yCos << SHIFT) >> randomTab(4),
                                                    (xSin << SHIFT) >> randomTab(4),
                                                    randomTab(100),
                                                    MAXY,
                                                    randomTab(trailLen) + 5,
                                                    colour);
                        }
                     }
                     break;

                case 6:
                     {
                        // ????????.
                        int r;
                        long xSin;
                        long yCos;

                        for (count = 0; count < starBurstNum; count++)
                        {
                            r    = randomTab(NUMOFDEGREES);
                            xSin = SIN (r);
                            yCos = COS (r);
                            pStarBurst[count].init (x,
                                                    y,
                                                    (yCos * randomTab(SHIFT)),
                                                    (xSin * randomTab(SHIFT)),
                                                    randomTab(100),
                                                    MAXY,
                                                    randomTab(trailLen) + 5,
                                                    colour);
                        }
                     }
                     break;

                case 7:
                     {
                        // ????????.
                        int r;
                        long xSin;
                        long yCos;

                        for (count = 0; count < starBurstNum; count++)
                        {
                            r    = randomTab(NUMOFDEGREES);
                            xSin = SIN (r) * 2;
                            yCos = COS (r) * 4;
                            pStarBurst[count].init (x,
                                                    y,
                                                    (yCos * randomTab(30 / 4)),
                                                    (xSin * randomTab(30 / 2)),
                                                    randomTab(100),
                                                    MAXY,
                                                    randomTab(trailLen) + 5,
                                                    colour);
                        }
                     }
                     break;



            }

            // To play sound we need to unlock screen, play sound, then
            // lock it again .. why ?
            if (ExplodeSnd != 0)
            {
            	ddObj.UnlockPrimarySurface();

	            switch (randomTab(500))
    	        {
        	    	case 255:
				    	PlaySound ("explode2.wav", NULL, SND_ASYNC);
                	    break;

	            	default:
				    	PlaySound ("explode1.wav", NULL, SND_ASYNC);
        	    }

	            ddObj.LockPrimarySurface();
            }

        }
    }
    else // Update flare explosion.
    {
        int aliveCount = starBurstNum;

        // Update and text each flare in the explosion, if all dead,
        // tell system.
        for (int count = 0; count < starBurstNum; count++)
        {

             if (pStarBurst[count].move(ddObj) == FALSE)
                aliveCount--;
             else if (ddObj.bitDepth > 8)
             {
                  short r = GetRValue(pStarBurst[count].colour);
                  short g = GetGValue(pStarBurst[count].colour);
                  short b = GetBValue(pStarBurst[count].colour);

                  switch (colItemChange)
                  {
                        // Red
                        case 0:
                        	 if (r > 0 && r < 255)
	                             r += (short) colChangeAmnt;
                             break;

                        // Green
                        case 1:
                        	 if (g > 0 && g < 255)
	                             g += (short) colChangeAmnt;
                             break;

                        // Blue
                        case 2:
                        	 if (b > 0 && b < 255)
                            	 b += (short) colChangeAmnt;
                             break;
                  }

                  pStarBurst[count].colour = RGB(r, g, b);
             }
        }

        // Return false if all flares are DEAD.
        if (aliveCount == 0)
           return FALSE;
    }

    return TRUE;
}

// ############################################################################
// ############################################################################
// ############################################################################

void MultiRocket::init (int  starBNum,   int      subStarBNum,
                        int  x,          int      y,
                        int  thr,        int      grav,
                        int  cyc,        int      gnd,
                        int trailLength, COLORREF col)
{
    //**** Allocate explode memory ****
    // If memory is already allocated, but explode number has changed. Then
    // delete/free memory.
    if ((pStarBurst != NULL) && (starBurstNum != starBNum))
    {
        delete[] pStarBurst;
        pStarBurst = NULL;
    }

    if (pStarBurst == NULL)
       pStarBurst = new SkyRocket[starBNum];

    // Initialise main rocket ....
    mainRocket.init (x, y, thr, grav, cyc, gnd, trailLength, col);

    // Initialise explode variables.
    explode          = FALSE;
    starBurstNum     = starBNum;
    subStartBurstNum = subStarBNum,
    colour           = col;
    trailLen         = trailLength;

    // Select the increment style according to current colour value.
    switch ((colItemChange = randomTab(3)))
    {
         // red
         case 0:
              if (GetRValue(colour) < 127)
                 colChangeAmnt = 1;
              else
                 colChangeAmnt = -1;
              break;

         // green
         case 1:
              if (GetGValue(colour) < 127)
                 colChangeAmnt = 1;
              else
                 colChangeAmnt = -1;
              break;


         // blue
         case 2:
             if (GetBValue(colour) < 127)
                colChangeAmnt = 1;
             else
                colChangeAmnt = -1;
             break;
    }
}

BOOL MultiRocket::move (DDWrapper& ddObj)
{
    // If no memory allocated .... leave.
    if (pStarBurst == NULL)
       return FALSE;

    // Update main rocket, and then explode!
    if (explode == FALSE)
    {
        if (mainRocket.move(ddObj) == FALSE)
        {
            // Set up explosion when main rocket dies.
            int x, y;
            explode = TRUE;

            // intialise star burst.
            mainRocket.getCurrentPos (x, y);

			int count;
            int r = randomTab (BOOMNUM); // make sure
            for (count = 0; count < starBurstNum; count++)
            {
	            pStarBurst[count].init (subStartBurstNum,
                                        x,
                                        y,
                                        ((randomTab (10) - 5) << SHIFT),
                                        (randomTab (20) << SHIFT),
                                        randomTab(20),
                                        MAXY,
                                        randomTab(trailLen)+5,
                                        colour);
                pStarBurst[count].multiBang = r;
            }
        }
    }
    else // Update flare explosion.
    {
        int aliveCount = starBurstNum;

        // Update and text each flare in the explosion, if all dead,
        // tell system.
        for (int count = 0; count < starBurstNum; count++)
        {

             if (pStarBurst[count].move(ddObj) == FALSE)
                aliveCount--;
             else if (ddObj.bitDepth > 8)
             {
                  short r = GetRValue(pStarBurst[count].colour);
                  short g = GetGValue(pStarBurst[count].colour);
                  short b = GetBValue(pStarBurst[count].colour);

                  switch (colItemChange)
                  {
                        // Red
                        case 0:
                        	 if (r > 0 && r < 255)
	                             r += (short) colChangeAmnt;
                             break;

                        // Green
                        case 1:
                        	 if (g > 0 && g < 255)
	                             g += (short) colChangeAmnt;
                             break;

                        // Blue
                        case 2:
                        	 if (b > 0 && b < 255)
                            	 b += (short) colChangeAmnt;
                             break;
                  }

                  pStarBurst[count].setColour(RGB(r, g, b));
             }
        }

        // Return false if all flares are DEAD.
        if (aliveCount == 0)
           return FALSE;
    }

    return TRUE;
}

// ############################################################################
// ############################################################################
// ############################################################################

void FlowerPot::init (int num, int x, int y, COLORREF col, int cyc, int tLen)
{
    if (num <= 0)
       return;

    // Allocate memory.
    if ((pFlares != NULL) && (num != flareNum))
    {
        delete[] pFlares;
        pFlares = NULL;
    }

    if (pFlares == NULL)
       pFlares = new Flare[num];

    flareNum = num;
    colour   = col;
    cycles   = cyc;
    xPos     = x;
    yPos     = y;
    trailLen = tLen;

    // Select the increment style according to current colour value.
    switch ((colItemChange = randomTab(3)))
    {
         // red
         case 0:
              if (GetRValue(colour) < 127)
                 colChangeAmnt = 1;
              else
                 colChangeAmnt = -1;
              break;

         // green
         case 1:
              if (GetGValue(colour) < 127)
                 colChangeAmnt = 1;
              else
                 colChangeAmnt = -1;
              break;


         // blue
         case 2:
             if (GetBValue(colour) < 127)
                colChangeAmnt = 1;
             else
                colChangeAmnt = -1;
             break;
    }

    // Setup inital flares !
    if (pFlares == NULL)
        return;

    for (int count = 0; count < flareNum; count++)
    {
         pFlares[count].init (xPos,
                              yPos,
                              ((randomTab (10) - 5) << SHIFT) >> (randomTab(3) + 1),
                              (randomTab (20) << SHIFT) >> (randomTab(2) + 1),
                              randomTab(50),
                              yPos,
                              trailLen,
                              colour);
    }
}

BOOL FlowerPot::move (DDWrapper& ddObj)
{
    // If noting allocated, nothing to move.
    if (pFlares == NULL)
       return FALSE;

    if (cycles > 0)
    {
        cycles--;
        for (int count = 0; count < flareNum; count++)
        {
            // Update flares
            if (pFlares[count].move(ddObj) == FALSE)
            {
                pFlares[count].init (xPos,
                                     yPos,
                                     ((randomTab (10)-5) << SHIFT) >> (randomTab(3) + 1),
                                     (randomTab (20) << SHIFT) >> (randomTab(2) + 1),
                                     randomTab(50),
                                     yPos,
                                     trailLen,
                                     colour);
            }
            else if (ddObj.bitDepth > 8)
            {
                  int r = GetRValue(pFlares[count].colour);
                  int g = GetGValue(pFlares[count].colour);
                  int b = GetBValue(pFlares[count].colour);

                  switch (colItemChange)
                  {
                        // Red
                        case 0:
                             r += colChangeAmnt;
                             break;

                        // Green
                        case 1:
                             g += colChangeAmnt;
                             break;

                        // Blue
                        case 2:
                             b += colChangeAmnt;
                             break;
                  }

                  pFlares[count].colour = RGB(r, g, b);
             }
        }
    }
    else
    {
        int testAlive = 0;

        for (int count = 0; count < flareNum; count++)
        {
            if (pFlares[count].move(ddObj) == FALSE)
                   testAlive++;
            else
            {
                pFlares[count].colour = RGB(GetRValue(pFlares[count].colour) -1,
                                            GetGValue(pFlares[count].colour) -1,
                                            GetBValue(pFlares[count].colour) -1);
            }
        }

        // Test if flares are still active.
        if (testAlive == flareNum)
            return FALSE;
        else
            return TRUE;
    }

    return TRUE;
}

// ############################################################################

#endif // _FCLASSES_CODE
