#ifndef _FCLASSES_HEADER
#define _FCLASSES_HEADER

#include <windows.h>
#include <mmsystem.h>
#include <stdlib.h>
#include <time.h>
#include "DDWrapper.h"
#include "fixed_math.h"

// **** Fire work constants ****
extern int MAXX, MAXY;
extern int PixelSize;
extern int ExplodeSnd;

// used int flare trailer to save previous points in a queue structure
struct SaveXY
{
    int x, y;
};

class Flare
{
   public:
            int      xPos, yPos;                       // start position of firework.
            int      ground, cycles;                   // ground limit, and cycles of life for flare.
            int      xPlot, yPlot;                     // X, Y increments for firework (fixed point).
            int      gravConst, gravty, thrust;        // gravity pull on flare, virtical thrust on gravity, horizontal thrust
            SaveXY*  trailer;                          // point to trailer queue structure
            int      trailLen, bufPnt;                 // current size of trailer, current point in trailer element.
            COLORREF colour;                           // current colour of flare.
            BOOL     overBuf, alive;                   // over write trailer buffer, and determine in flare is still active.
            BOOL     gndTerminated;                    // Test flare has hit the ground, if so then don't update x, y movement!

   protected:

            // Removes a point from the display, also adjust flooring of flare if
            // it's below the floor limit.
            void removePoint (DDWrapper& ddObj, int x, int y);

            // Displays a point from the display, also adjust flooring of flare if
            // it's below the floor limit.
            void showPoint (DDWrapper& ddObj, int x, int y);

   public:
           // ********* Constructors **********
           Flare         (void);
           Flare         (int  x,                  int      y,
                          int thr = (1 << SHIFT),  int      grav = (50 << SHIFT),
                          int  cyc = 20,           int      gnd  = MAXY,
                          int  trailLength = 5,    COLORREF col  = RGB(255,255,255));

           // Setup a flare.
           // x, y    :        Start location on the screen.
           // thr     :        Horizontal pull of flare.
           // gnd     :        Virtual pull of flare.
           // cyc     :        Number of iterations of life (total life = cyc+trailLength).
           // gnd     :        Set floor limit.
           // trailLength:     Length of flare trail.
           // col     :        Set colour of flare.
           void  init    (int  x,                  int      y,
                          int thr = (1 << SHIFT),  int     grav = (50 << SHIFT),
                          int  cyc = 20,           int      gnd  = MAXY,
                          int  trailLength = 5,    COLORREF col  = RGB(255,255,255));

           // This function will remove, move, and display a flare.
           // ddObj  :    direct draw object.
           // returns true/false if flare is active.
           BOOL  move     (DDWrapper& ddObj);

           // Returns the current position of the flare
           void  getCurrentPos (int& x, int& y)
           {
               x = xPos + (xPlot >> SHIFT);
               y = yPos - (yPlot >> SHIFT);
           }

        // ********* Destructor **********
        virtual  ~Flare   (void);
};


#define BOOMNUM 8 // number of different explosions.
class SkyRocket
{
    public:
            Flare    mainRocket;        // Main rocket object fire into the sky/screen
            Flare*   pStarBurst;        // Pointer to an array of flare objects.
            BOOL     explode;           // Variable used to determine if updating main rocket or explosion.
            int      starBurstNum;      // Num of flares in explosion.
            int      trailLen;          // Trailer len of explosions
            COLORREF colour;            // Colour sky rocket.

            // Variables used to calculate the colour change during the
            // life of flare!
            int      colItemChange;     // R, G, B value to change
            int      colChangeAmnt;     // Positive/Negative amount to change.

            int	 	 multiBang;
    public:
           // **** constructor ****
           SkyRocket ()
           {
               starBurstNum = 0;
               pStarBurst   = NULL;
               explode      = FALSE;
           }

           // Setup a skyrocket.
           // starBNum:        Number of objects in explosion.
           // x, y    :        Start location on the screen.
           // thr     :        Horizontal pull of flare.
           // gnd     :        Virtual pull of flare.
           // cyc     :        Number of iterations of life (total life = cyc+trailLength).
           // gnd     :        Set floor limit.
           // trailLength:     Length of flare trail.
           // col     :        Set colour of flare.
           void init      (int  starBNum,
                           int  x,                  int      y,
                           int thr = (1 << SHIFT),  int      grav = (50 << SHIFT),
                           int  cyc = 20,           int      gnd  = MAXY,
                           int  trailLength = 5,    COLORREF col  = RGB(255,255,255));


           // This function will remove, move, and display a flare.
           // ddObj  :    direct draw object.
           // returns true/false if flare is active.
           BOOL  move     (DDWrapper& ddObj);

           COLORREF getColour (void)
           {
                return mainRocket.colour;
           }

           void  setColour(COLORREF col)
           {
                mainRocket.colour = col;
           }


           // **** Destructor ****
           ~SkyRocket()
           {
				if (pStarBurst != NULL)
                {
                	delete[] pStarBurst;
                    pStarBurst = NULL;
                }
           }
};

// ########################################################################################

class MultiRocket
{
    public:
            Flare      mainRocket;        // Main rocket object fire into the sky/screen
            SkyRocket* pStarBurst;        // Pointer to an array of flare objects.
            BOOL       explode;           // Variable used to determine if updating main rocket or explosion.
            int        starBurstNum;      // Num of flares in explosion.
            int		   subStartBurstNum;  // Num of sub-flares to explode.
            int        trailLen;          // Trailer len of explosions
            COLORREF   colour;            // Colour sky rocket.

            // Variables used to calculate the colour change during the
            // life of flare!
            int      colItemChange;     // R, G, B value to change
            int      colChangeAmnt;     // Positive/Negative amount to change.

    public:
           // **** constructor ****
           MultiRocket ()
           {
               starBurstNum = 0;
               pStarBurst   = NULL;
               explode      = FALSE;
           }

           // Setup a skyrocket.
           // starBNum:        Number of objects in explosion.
		   // subStarBNum
           // x, y    :        Start location on the screen.
           // thr     :        Horizontal pull of flare.
           // gnd     :        Virtual pull of flare.
           // cyc     :        Number of iterations of life (total life = cyc+trailLength).
           // gnd     :        Set floor limit.
           // trailLength:     Length of flare trail.
           // col     :        Set colour of flare.
           void init      (int  starBNum, 		    int      subStarBNum,
                           int  x,                  int      y,
                           int thr = (1 << SHIFT),  int      grav = (50 << SHIFT),
                           int  cyc = 20,           int      gnd  = MAXY,
                           int  trailLength = 5,    COLORREF col  = RGB(255,255,255));


           // This function will remove, move, and display a flare.
           // ddObj  :    direct draw object.
           // returns true/false if flare is active.
           BOOL  move     (DDWrapper& ddObj);

           // **** Destructor ****
           ~MultiRocket()
           {
				if (pStarBurst != NULL)
                {
                	delete[] pStarBurst;
                    pStarBurst = NULL;
                }
           }
};

// ############################################################################################

class FlowerPot
{
    private:
            Flare*   pFlares;     // Point to flare object list.
            int      flareNum;    // Flare number to allocate.
            COLORREF colour;      // Inital colour of flares.
            int      cycles;      // Cycles of life for flower pot.
            int      xPos, yPos;  // Inital x, y, pos of flower pot
            int      trailLen;    // Trailer length variable.

            // Variables used to calculate the colour change during the
            // life of flare!
            int      colItemChange;     // R, G, B value to change
            int      colChangeAmnt;     // Positive/Negative amount to change.

    public:
           // **** constructor ****
           FlowerPot ()
           {
                pFlares = NULL;
                xPos = yPos = cycles = colour = flareNum = 0;
           }

           // Initalise the flower pot.
           // num       : Number of flares to display.
           // x,y       : Start position of flower pot.
           // col       : Colour of flower pot.
           // cyc       : How long flower pot will live.
           // tLen      : Trailer length
           void init (int num, int x, int y, COLORREF col, int cyc, int tLen);

           // This function will remove, move, and display a flare.
           // ddObj  :    direct draw object.
           // returns true/false if flare is active.
           BOOL  move     (DDWrapper& ddObj);

           // **** Destructor ****
           ~FlowerPot()
           {
				if (pFlares != NULL)
                {
                	delete[] pFlares;
                    pFlares = NULL;
                }
           }
};

// ############################################################################################

#endif // _FCLASSES_HEADER
